#  tests for docker-py-4.4.1-py310h9f7ea03_5 (this is a generated file);
print('===== testing package: docker-py-4.4.1-py310h9f7ea03_5 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
#!/usr/bin/env python

from shutil import which
from sys import exit

if which('docker'):
    try:
        import docker
        docker.DockerClient.from_env(version='auto')
        print("INFO :: docker.DockerClient.from_env(version='auto') :: PASSED")
    except:
        print("ERROR :: docker.DockerClient.from_env(version='auto') :: FAILED")
        exit(1)
    try:
        import pkg_resources
        pkg_resources.require('docker')
        print("INFO :: pkg_resources.require('docker') :: PASSED")
    except:
        print("INFO :: pkg_resources.require('docker') :: FAILED")
        exit(1)
else:
    print("WARNING :: which('docker') failed")
    print("WARNING :: docker.DockerClient.from_env() and")
    print("WARNING :: pkg_resources.require('docker') tests skipped")
exit(0)
#  --- run_test.py (end) ---

print('===== docker-py-4.4.1-py310h9f7ea03_5 OK =====');
print("import: 'docker'")
import docker

print("import: 'docker.api'")
import docker.api

print("import: 'docker.auth'")
import docker.auth

print("import: 'docker.tls'")
import docker.tls

print("import: 'docker.utils'")
import docker.utils

print("import: 'docker.utils.ports'")
import docker.utils.ports

print("import: 'docker.transport.unixconn'")
import docker.transport.unixconn

print("import: 'docker.transport.npipeconn'")
import docker.transport.npipeconn

print("import: 'docker.transport.npipesocket'")
import docker.transport.npipesocket

